from typing import TypeVar, Callable, Generic, Tuple

from amino import do, Do, List, Just
from amino.case import Case
from amino.logging import module_log
from amino.lenses.lens import lens

from ribosome.util.menu.data import (Menu, MenuAction, MenuQuit, MenuPrompt, MenuUpdateLines, MenuUnit, MenuState,
                                     MenuUpdateCursor, visible_lines, MenuStackAction, MenuStackQuit, MenuStackPush,
                                     MenuStackPop, Menus, MenuPush, MenuPop, MenuQuitWith, MenuConfig)
from ribosome.nvim.io.compute import NvimIO
from ribosome.util.menu.prompt.run import prompt
from ribosome.nvim.scratch import (create_scratch_buffer, CreateScratchBufferOptions, ScratchBuffer,
                                   set_scratch_buffer_content)
from ribosome.util.menu.prompt.data import (InputChar, InputState, PromptAction, PromptUnit, PromptStateTrans,
                                            PromptUpdate, PromptQuit)
from ribosome.nvim.io.state import NS
from ribosome.nvim.api.ui import set_cursor, close_buffer
from ribosome.compute.prog import Prog
from ribosome.compute.ribosome_api import Ribo
from ribosome.nvim.io.api import N

log = module_log()
S = TypeVar('S')
ML = TypeVar('ML')
U = TypeVar('U')


@do(NvimIO[None])
def cleanup_menu(scratch: ScratchBuffer) -> Do:
    yield N.recover_failure(close_buffer(scratch.buffer), lambda a: N.unit)


@do(NS[InputState[MenuState[S, ML], U], PromptAction])
def menu_reentry(config: MenuConfig[S, ML, U], scratch: ScratchBuffer) -> Do:
    state = yield NS.get()
    return NS.lift(prompt(update_menu(config, scratch), state))


class execute_menu_action(Case[MenuAction, NS[InputState[MenuState[S, ML], U], PromptAction]], alg=MenuAction):

    def __init__(self, scratch: ScratchBuffer, config: MenuConfig[S, ML, U]) -> None:
        self.scratch = scratch
        self.config = config

    def quit(self, action: MenuQuit) -> NS[InputState[MenuState[S, ML], U], PromptAction]:
        return NS.pure(PromptQuit())

    @do(NS[InputState[MenuState[S, ML], U], PromptAction])
    def quit_with(self, action: MenuQuitWith) -> Do:
        yield NS.modify(lens.data.result.set(Just(action.next)))
        return PromptQuit()

    def prompt_state_trans(self, action: MenuPrompt) -> NS[InputState[MenuState[S, ML], U], PromptAction]:
        return NS.pure(PromptStateTrans(action.state))

    @do(NS[InputState[MenuState[S, ML], U], PromptAction])
    def update_lines(self, action: MenuUpdateLines) -> Do:
        yield NS.lift(set_scratch_buffer_content(self.scratch, visible_lines(action.content).map(lambda a: a.text)))
        return PromptUnit()

    @do(NS[InputState[MenuState[S, ML], U], PromptAction])
    def update_cursor(self, a: MenuUpdateCursor) -> Do:
        line = yield NS.inspect(lambda a: a.data.cursor)
        yield NS.lift(set_cursor(self.scratch.ui.window, (line + 1, 0)))
        return PromptUnit()

    @do(NS[InputState[MenuState[S, ML], U], PromptAction])
    def push(self, action: MenuPush) -> Do:
        current = yield menu_reentry(self.config, self.scratch)
        yield NS.modify(lens.data.next.set(MenuStackPush(current, action.thunk(self.scratch))))
        return PromptQuit()

    @do(NS[InputState[MenuState[S, ML], U], PromptAction])
    def pop(self, action: MenuPop) -> Do:
        yield NS.modify(lens.data.next.set(MenuStackPop()))
        return PromptQuit()

    def unit(self, a: MenuUnit) -> NS[InputState[MenuState[S, ML], U], PromptAction]:
        return NS.pure(PromptUnit())


def update_menu(
        config: MenuConfig[S, ML, U],
        scratch: ScratchBuffer,
) -> Callable[[InputChar], NS[InputState[MenuState[S, ML], U], PromptAction]]:
    @do(NS[InputState[MenuState[S, ML], U], PromptAction])
    def update_menu(update: PromptUpdate[U]) -> Do:
        action = yield config.handle_input(update)
        yield execute_menu_action(scratch, config)(action)
    return update_menu


class menu_recurse(Generic[S, ML], Case[MenuStackAction, NS[Menus, Prog[None]]], alg=MenuStackAction):

    def __init__(self, result: MenuState[S, ML]) -> None:
        self.result = result

    def quit(self, action: MenuStackQuit) -> NS[Menus, Prog[None]]:
        return NS.pure(self.result.result)

    @do(NS[Menus, Prog[None]])
    def push(self, action: MenuStackPush) -> Do:
        yield NS.modify(lambda a: a.mod.stack(lambda b: b.cons(action.current)))
        result = yield action.thunk
        yield menu_recurse(result)(result.next)

    @do(NS[Menus, Prog[None]])
    def pop(self, action: MenuStackPop) -> Do:
        stack = yield NS.inspect(lambda a: a.stack)
        def push(
                menus: Menus,
                h: NS[Menus, Prog[None]],
                t: List[Menus, Prog[None]],
        ) -> NvimIO[Tuple[Menus, NS[Menus, Prog[None]]]]:
            return N.pure((menus.set.stack(t), h))
        next = yield stack.detach_head.map2(
            lambda h, t: NS.apply(lambda a: push(a, h, t))
        ).get_or(NS.pure, NS.pure(self.result.result))
        yield next


@do(NS[Menus, Prog[None]])
def menu_loop(menu: Menu[S, ML, U], scratch: ScratchBuffer) -> Do:
    result = yield NS.lift(prompt(update_menu(menu.config, scratch), menu.state))
    yield menu_recurse(result)(result.next)


@do(NvimIO[Prog[None]])
def run_menu(menu: Menu[S, ML, U]) -> Do:
    scratch = yield create_scratch_buffer(CreateScratchBufferOptions.cons(wrap=False, name=menu.config.name))
    state, next = yield N.ensure(menu_loop(menu, scratch).run(Menus.cons()), lambda a: cleanup_menu(scratch))
    return next.get_or_strict(Prog.unit).replace(state)


@do(Prog[S])
def run_menu_prog(menu: Menu[S, ML, U]) -> Do:
    next = yield Ribo.lift_nvimio(run_menu(menu))
    yield next


__all__ = ('run_menu', 'run_menu_prog',)



from typing import Callable, Any

from kallikrein import Expectation

from amino import do, Do, IO, Lists
from amino.logging import module_log
from amino.string.hues import green

from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.io.api import N
from ribosome.nvim.api.rpc import nvim_quit
from ribosome.nvim.io.data import NResult
from ribosome.test.config import TestConfig
from ribosome.test.integration.rpc import setup_test_nvim_embed, set_nvim_vars
from ribosome.test.run import run_test_io, plugin_started
from ribosome.test.integration.start import start_plugin_embed

log = module_log()


def cleanup(config: TestConfig) -> Callable[[NResult], NvimIO[None]]:
    @do(NvimIO[None])
    def cleanup(result: NResult) -> Do:
        yield nvim_quit()
        runtime_log = yield N.from_io(IO.delay(config.log_file.read_text))
        if runtime_log:
            log.info(green('plugin output:'))
            Lists.lines(runtime_log) % log.info
            log.info('')
    return cleanup


def plugin_test(config: TestConfig, io: Callable[..., NvimIO[Expectation]], *a: Any, **kw: Any) -> Expectation:
    @do(NvimIO[None])
    def run_nvim_io() -> Do:
        yield set_nvim_vars(config.vars + ('components', config.components))
        yield config.pre()
        if config.autostart:
            yield start_plugin_embed(config)
            yield plugin_started()
        yield io(*a, **kw)
    @do(IO[Expectation])
    def run() -> Do:
        nvim = yield setup_test_nvim_embed(config)
        yield N.to_io_a(N.ensure(run_nvim_io(), cleanup(config)), nvim.api)
    return run_test_io(run)


__all__ = ('plugin_test',)



from amino.io import IOExceptionBase
from amino import Maybe, List, Just
from amino.util.trace import default_internal_packages


class NvimIOException(IOExceptionBase):

    @property
    def desc(self) -> str:
        return 'NvimIO exception'

    @property
    def internal_packages(self) -> Maybe[List[str]]:
        return Just(default_internal_packages.cons('ribosome.nvim'))


__all__ = ('NvimIOException',)



from amino.case import Case
from amino import do, Do, Dat, Just, Nothing
from amino.state import State
from amino.util.string import camelcase

from ribosome.util.doc.data import Anchor, VariableAnchor, AnchorType, MappingAnchor, RpcAnchor, GeneralAnchor
from ribosome.rpc.data.prefix_style import Short, Full


class CompilerConfig(Dat['CompilerConfig']):

    @staticmethod
    def cons(
            name: str,
            prefix: str=None,
    ) -> 'CompilerConfig':
        return CompilerConfig(
            name,
            prefix or name,
        )

    def __init__(self, name: str, prefix: str) -> None:
        self.name = name
        self.prefix = prefix


def format_variable(name: str, anchor: Anchor, tpe: VariableAnchor) -> str:
    prefix = f'{name}_' if tpe.prefix else ''
    return f'{tpe.scope}:{prefix}{anchor.text}'


class compile_anchor_type(Case[AnchorType, State[CompilerConfig, str]], alg=AnchorType):

    def __init__(self, anchor: Anchor) -> None:
        self.anchor = anchor

    @do(State[CompilerConfig, str])
    def variable(self, a: VariableAnchor) -> Do:
        name = yield State.inspect(lambda a: a.name)
        return format_variable(name, self.anchor, a)

    def mapping(self, a: MappingAnchor) -> State[CompilerConfig, str]:
        return State.pure('')

    @do(State[CompilerConfig, str])
    def rpc(self, a: RpcAnchor) -> Do:
        name = yield State.inspect(lambda a: a.name)
        prefix = yield State.inspect(lambda a: a.prefix)
        rpc_prefix = Just(name) if a.prefix == Full() else Just(prefix) if a.prefix == Short() else Nothing
        prefix_s = rpc_prefix.map(lambda a: f'{a}_').get_or_strict('')
        cc = camelcase(f'{prefix_s}{self.anchor.text}')
        return f':{cc}'

    def general(self, a: GeneralAnchor) -> State[CompilerConfig, str]:
        return State.pure(self.anchor.text)


def compile_anchor(anchor: Anchor) -> State[CompilerConfig, str]:
    return compile_anchor_type(anchor)(anchor.tpe)


__all__ = ('format_variable', 'compile_anchor',)



import abc
import json
from typing import Callable, Type, TypeVar, Generic, Any, Union

from amino import List, Either, __, Left, Eval, ADT, Right, Try, Path, Map, Lists
from amino.do import do, Do
from amino.boolean import false, true
from amino.json.decoder import decode_json_type

from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.api.variable import variable_prefixed, variable, variable_set, variable_set_prefixed
from ribosome.nvim.api.util import cons_checked_e
from ribosome.nvim.io.api import N
from ribosome.util.doc.data import DocBlock

A = TypeVar('A', contravariant=True)
B = TypeVar('B')


class Setting(Generic[B], ADT['Setting[B]']):

    @abc.abstractproperty
    def value(self) -> NvimIO[Either[str, B]]:
        ...

    @abc.abstractproperty
    def default_e(self) -> Either[str, B]:
        ...

    @abc.abstractmethod
    def update(self, value: B) -> NvimIO[None]:
        ...

    @do(NvimIO[None])
    def ensure(self, fallback: B) -> Do:
        current = yield self.value
        yield current.cata(lambda e: self.update(fallback), lambda a: N.unit)

    @do(NvimIO[Either[str, B]])
    def value_or_default_e(self) -> Do:
        value = yield self.value
        return value.o(self.default_e)

    @do(NvimIO[B])
    def value_or_default(self) -> Do:
        value = yield self.value_or_default_e()
        yield N.from_either(value)


class StrictSetting(Generic[A, B], Setting[B]):

    def __init__(
            self,
            name: str,
            desc: str,
            help: str,
            prefix: bool,
            tpe: Type[A],
            ctor: Callable[[A], Either[str, B]],
            default: Either[str, B],
    ) -> None:
        self.name = name
        self.desc = desc
        self.help = help
        self.prefix = prefix
        self.tpe = tpe
        self.ctor = ctor
        self.default = default

    def _arg_desc(self) -> List[str]:
        return List(self.name, str(self.prefix), str(self.tpe))

    @property
    def value(self) -> NvimIO[Either[str, B]]:
        api_call = variable_prefixed if self.prefix else variable
        return api_call(self.name, cons_checked_e(self.tpe, self.ctor))

    def value_or(self, default: B) -> NvimIO[B]:
        return self.value / __.get_or_else(default)

    @property
    def default_e(self) -> Either[str, B]:
        return self.default

    def update(self, value: B) -> NvimIO[None]:
        api_call = variable_set_prefixed if self.prefix else variable_set
        return api_call(self.name, value)


class EvalSetting(Generic[B], Setting[B]):

    def __init__(
            self,
            name: str,
            f: Eval[NvimIO[Either[str, B]]],
            default: Either[str, B]=Left('no default specified')
    ) -> None:
        self.name = name
        self.f = f
        self.default = default

    def _arg_desc(self) -> List[str]:
        return List(self.name)

    @property
    def value(self) -> NvimIO[Either[str, B]]:
        return self.f.value

    @property
    def default_e(self) -> Either[str, B]:
        return self.default

    def update(self, value: B) -> NvimIO[None]:
        return N.pure(None)


SettingCtor = Callable[[str, str, str, bool, Either[str, B]], Setting[B]]


def no_default(name: str) -> Either[str, A]:
    return Left(f'no default specified for setting `{name}`')


def setting_ctor(tpe: Type[A], ctor: Callable[[A], Either[str, B]]) -> SettingCtor:
    def setting(
            name: str,
            desc: str,
            help: Union[str, DocBlock],
            prefix: bool,
            default: Either[str, B]=None,
    ) -> Setting[B]:
        return StrictSetting(
            name,
            desc,
            DocBlock.string(help) if isinstance(help, str) else help,
            prefix,
            tpe,
            ctor,
            default or no_default(name),
        )
    return setting


def path_list(data: list) -> Either[str, List[Path]]:
    return Lists.wrap(data).traverse(lambda a: Try(Path, a) / __.expanduser(), Either)


def str_list(data: list) -> Either[str, List[str]]:
    return Lists.wrap(data).traverse(lambda a: Right(a) if isinstance(a, str) else Left(f'not a string: {a}'), Either)


@do(Either[str, A])
def decode_json_setting(tpe: Type[A], data: Any) -> Do:
    js = yield Try(json.dumps, data)
    yield decode_json_type(js, tpe)


def json_setting(tpe: Type[A]) -> SettingCtor[A]:
    return setting_ctor(object, lambda a: decode_json_setting(tpe, a))


str_setting = setting_ctor(str, Right)
int_setting = setting_ctor(int, Right)
float_setting = setting_ctor(float, Right)
list_setting = setting_ctor(list, Right)
str_list_setting = setting_ctor(list, str_list)
path_setting = setting_ctor(str, (lambda a: Try(Path, a)))
path_list_setting = setting_ctor(list, path_list)
map_setting = setting_ctor(dict, lambda a: Right(Map(a)))
path_map_setting = setting_ctor(dict, lambda a: Try(Map, a).valmap(lambda b: Path(b).expanduser()))
bool_setting = setting_ctor(int, lambda a: Right(false if a == 0 else true))


__all__ = ('Setting', 'StrictSetting', 'EvalSetting', 'setting_ctor', 'str_setting', 'int_setting', 'float_setting',
           'list_setting', 'path_setting', 'path_list_setting', 'map_setting', 'path_map_setting', 'bool_setting',
           'str_list_setting', 'json_setting',)



from typing import TypeVar
from types import ModuleType
import pkgutil
import importlib

from amino import do, Do, IO, List, Either, Lists, Try, Just, Nil, Map, Maybe, Right
from amino.either import ImportFailure
from amino.mod import instances_from_module
from amino.dat import DatMeta, Field
from amino.logging import module_log

from ribosome.config.setting import StrictSetting
from ribosome.util.doc.data import (StaticDoc, DocCompiler, DocBlock, DocLine, DocString, Headline, Anchor,
                                    VariableAnchor, RpcAnchor, DocCat, Code)
from ribosome.config.component import Component
from ribosome.rpc.api import RpcProgram

log = module_log()
A = TypeVar('A')
B = TypeVar('B')


def json_param(field: Field, help: str) -> DocLine[A]:
    return DocLine(DocCat(List(
        DocString(f'{field.name}: {field.tpe}', Code('python')),
        DocString.none(help),
    )))


def json_params(tpe: DatMeta, help: Map[str, str]) -> Either[str, List[DocLine[A]]]:
    ps = tpe._dat__fields.traverse(lambda a: help.lift(a.name).map(lambda b: json_param(a, b)), Maybe)
    return ps.to_either(f'not all fields of {tpe} are documented')


@do(Either[str, List[DocBlock[A]]])
def rpc_params(rpc: RpcProgram) -> Do:
    params = rpc.program.metadata.params_spec
    json = yield (
        params.types.last.zip(rpc.options.json_help).map2(json_params).get_or_strict(Right(Nil))
        if rpc.options.json else
        Right(Nil)
    )
    return List(DocBlock.none(json))


@do(Either[str, List[DocBlock[A]]])
def rpc_doc(rpc: RpcProgram) -> Do:
    name = DocBlock.headline(rpc.program_name, 4, Anchor(rpc.program_name, RpcAnchor(rpc.options.prefix)))
    params = yield rpc_params(rpc)
    return List(name, rpc.options.help) + params.cat(DocBlock.empty())


@do(IO[List[DocBlock[A]]])
def component_doc(component: Component) -> Do:
    headline = DocBlock.headline(component.name, 2)
    desc = component.help.to_list
    rpc = yield IO.e(component.rpc.flat_traverse(rpc_doc, Either))
    rpc_blocks = rpc.cons(DocBlock.headline('RPC', 3))
    return List(headline) + desc + rpc_blocks.cat(DocBlock.empty())


@do(IO[List[DocBlock[A]]])
def components_doc(components: List[Component]) -> Do:
    main = yield components.flat_traverse(component_doc, IO)
    return main.cons(DocBlock.none(List(DocLine.headline('Components', 1))))


def setting_doc(setting: StrictSetting) -> IO[List[DocBlock[A]]]:
    anchor = Anchor(setting.name, VariableAnchor('g', setting.prefix))
    headline = DocLine(DocString(setting.name, Headline(2, Just(anchor))))
    desc = DocLine.string(setting.desc)
    block1: DocBlock[A] = DocBlock.none(List(headline, desc, DocLine.empty()))
    blocks = List(block1, setting.help, DocBlock.none(List(DocLine.empty())))
    return IO.pure(blocks)


@do(IO[List[DocBlock[A]]])
def settings_doc(settings: List[StrictSetting]) -> Do:
    main = yield settings.flat_traverse(setting_doc, IO)
    return main.cons(DocBlock.none(List(DocLine.headline('Settings', 1))))


@do(IO[List[str]])
def generate_doc(
        components: List[Component],
        settings: List[StrictSetting],
        static: StaticDoc,
        compiler: DocCompiler[A, B],
) -> Do:
    components_lines = yield components_doc(components)
    settings_lines = yield settings_doc(settings)
    yield compiler.compile(static.pre + components_lines + settings_lines + static.post, compiler.conf)


def submodules(mod: ModuleType) -> List[str]:
    return Lists.wrap(pkgutil.iter_modules(mod.__path__)).map(lambda a: f'{mod.__name__}.{a.name}')


def import_mods(mods: List[str]) -> Either[str, List[ModuleType]]:
    return mods.traverse(lambda a: Try(importlib.import_module, a), Either)


@do(Either[ImportFailure, List[Component]])
def resolve_components(module: str) -> Do:
    base = yield Try(importlib.import_module, module)
    component_names = submodules(base)
    component_pkgs = yield import_mods(component_names)
    component_mod_names = component_pkgs.flat_map(submodules)
    component_modules = yield import_mods(component_mod_names)
    modules = component_modules + component_pkgs
    return modules.flat_map(lambda a: instances_from_module(a, Component)).join


@do(IO[List[str]])
def generate_plugin_doc(
        components_module: str,
        settings_modules: List[ModuleType],
        static: StaticDoc,
        compiler: DocCompiler[A, B],
) -> Do:
    components = yield IO.e(resolve_components(components_module))
    settings = yield IO.e(settings_modules.flat_traverse(lambda a: instances_from_module(a, StrictSetting), Either))
    yield generate_doc(components, settings, static, compiler)


__all__ = ('generate_doc', 'generate_plugin_doc', 'setting_doc', 'settings_doc',)



from typing import TypeVar, Any, Generic

from amino import List, _, __
from amino.do import do, Do
from amino.case import Case

from ribosome.nvim.io.state import NS
from ribosome.compute.prog import Prog, ProgBind, ProgPure, ProgError, ProgExec
from ribosome.compute.wrap_data import ProgWrappers
from ribosome.data.plugin_state import PluginState
from ribosome.compute.program import bind_program, Program
from ribosome.compute.interpret import interpret

A = TypeVar('A')
B = TypeVar('B')
D = TypeVar('D')
R = TypeVar('R')
CC = TypeVar('CC')


@do(NS[PluginState[D, CC], A])
def transform_prog_state(st: NS[R, A], wrappers: ProgWrappers[PluginState[D, CC], R]) -> Do:
    yield st.transform_s(wrappers.get, wrappers.put)


def log_prog(prog: ProgExec) -> NS[PluginState[D, CC], None]:
    return NS.pure(None) if prog.name in ('program_log', 'pure', 'lift') else NS.modify(__.log_prog(prog.name))


class eval_prog(Generic[A, B, R, D, CC], Case[Prog[A], NS[PluginState[D, CC], A]], alg=Prog):

    @do(NS[PluginState[D, CC], A])
    def prog_exec(self, prog: ProgExec[B, A, R, Any]) -> Do:
        yield log_prog(prog)
        io_interpreter = yield NS.inspect(_.io_interpreter)
        output = yield transform_prog_state(prog.code, prog.wrappers)
        yield self(interpret(io_interpreter)(prog.output_type, output))

    @do(NS[PluginState[D, CC], A])
    def prog_bind(self, prog: ProgBind[Any, A]) -> Do:
        result = yield self(prog.fa)
        yield self(prog.f(result))

    @do(NS[PluginState[D, CC], A])
    def prog_pure(self, prog: ProgPure[A]) -> Do:
        yield NS.pure(prog.value)

    @do(NS[PluginState[D, CC], A])
    def prog_error(self, prog: ProgError[A]) -> Do:
        yield NS.error(prog.msg)


@do(NS[PluginState[D, CC], A])
def run_prog(program: Program[A], args: List[Any]) -> Do:
    yield eval_prog.match(bind_program(program, args))


__all__ = ('run_prog',)



from __future__ import annotations
import logging
from typing import TypeVar, Generic, Callable, Any, Type

from amino import Map, List, Nil, Either, _, Maybe, __
from amino.dat import Dat
from amino.util.string import camelcase
from amino.logging import module_log

from ribosome.config.component import Component, Components, NoComponentData, ComponentConfig
from ribosome.nvim.io.state import NS
from ribosome.compute.program import Program
from ribosome.config.resources import Resources
from ribosome.config.basic_config import BasicConfig
from ribosome.compute.output import ProgIO
from ribosome.compute.prog import Prog
from ribosome.compute.interpret import interpret_io, no_interpreter
from ribosome.rpc.define import ActiveRpcTrigger
from ribosome.rpc.api import RpcProgram

A = TypeVar('A')
C = TypeVar('C')
CC = TypeVar('CC')
CD = TypeVar('CD')
D = TypeVar('D')
DIO = TypeVar('DIO')
log = module_log()


def component_ctor(comp: Component[CD, CC]) -> Callable[[], CD]:
    return comp.state_ctor.get_or_strict(NoComponentData)


def component_ctor_m(comp: Either[str, Component[CD, CC]]) -> Callable[[], CD]:
    return (comp.to_maybe // _.state_ctor).get_or_strict(NoComponentData)


# FIXME in order to allow transparent updating of components, the data and config part must be separated
# FIXME using state type as a key won't work if multiple comps have the same type
class PluginState(Generic[D, CC], Dat['PluginState[D, CC]']):

    @staticmethod
    def cons(
            basic: BasicConfig,
            comp: ComponentConfig,
            rpc: List[RpcProgram],
            data: D,
            components: List[Component],
            init: Program,
            program_log: List[str]=Nil,
            logger: Program[None]=None,
            log_handler: logging.Handler=None,
            component_data: Map[type, Any]=Map(),
            active_mappings: Map[str, Program]=Map(),
            io_executor: Callable[[DIO], NS['PluginState[D, CC]', Any]]=None,
            rpc_triggers: List[ActiveRpcTrigger]=Nil,
            programs: List[Program]=Nil,
            io_interpreter: Callable[[ProgIO], Prog]=None,
            custom_io: Callable[[Any], Prog[A]]=None,
    ) -> 'PluginState':
        return PluginState(
            basic,
            comp,
            rpc,
            data,
            Components.cons(components),
            init,
            program_log,
            Maybe.optional(log_handler),
            component_data,
            active_mappings,
            rpc_triggers,
            Maybe.optional(io_executor),
            programs,
            io_interpreter or interpret_io(custom_io or no_interpreter, Maybe.optional(logger)),
        )

    def __init__(
            self,
            basic: BasicConfig,
            comp: ComponentConfig,
            rpc: List[RpcProgram],
            data: D,
            components: Components,
            init: Program,
            program_log: List[str],
            log_handler: Maybe[logging.Handler],
            component_data: Map[type, Any],
            active_mappings: Map[str, Program],
            io_executor: Maybe[Callable[[DIO], NS['PluginState[D, CC]', Any]]],
            rpc_triggers: List[ActiveRpcTrigger],
            programs: List[Program],
            io_interpreter: Callable[[ProgIO], Prog],
    ) -> None:
        self.basic = basic
        self.comp = comp
        self.data = data
        self.components = components
        self.init = init
        self.program_log = program_log
        self.log_handler = log_handler
        self.component_data = component_data
        self.active_mappings = active_mappings
        self.io_executor = io_executor
        self.rpc_triggers = rpc_triggers
        self.programs = programs
        self.rpc = rpc
        self.io_interpreter = io_interpreter

    def update(self, data: D) -> 'PluginState[D, CC]':
        return self.copy(data=data)

    def log_prog(self, prog: str) -> 'PluginState[D, CC]':
        return self.append1.program_log(prog)

    def component(self, name: str) -> Either[str, Component]:
        return self.components.by_name(name)

    def ctor_by_type(self, tpe: Type[C]) -> Callable[[], C]:
        return component_ctor_m(self.components.by_type(tpe))

    def data_by_type(self, tpe: Type[C]) -> C:
        return self.component_data.lift(tpe) | (lambda: self.ctor_by_type(tpe)())

    def data_for(self, component: Component) -> Any:
        return self.component_data.lift(component.state_type).get_or(component_ctor(component))

    def data_by_name(self, name: str) -> Either[str, Any]:
        return self.component(name) / self.data_for

    def update_component_data(self, new: C) -> 'PluginState[D, CC]':
        return self.mod.component_data(__.cat((type(new), new)))

    def modify_component_data(self, name: str, mod: Callable[[Any], Any]) -> 'PluginState[D, CC]':
        comp = self.data_by_name(name)
        return comp / mod / self.update_component_data | self

    def resources_with(self, data: A) -> Resources[D, CC]:
        return Resources(data, self.components)

    @property
    def resources(self) -> Resources[D, CC]:
        return self.resources_with(self.data)

    @property
    def camelcase_name(self) -> str:
        return camelcase(self.basic.name)

    def programs_by_name(self, name: str) -> List[RpcProgram]:
        return self.programs.filter(_.rpc_name == name)

    def program_by_name(self, name: str) -> Either[str, RpcProgram]:
        return self.programs_by_name(name).head.to_either(f'no program named `{name}`')


PS = PluginState[D, CC]


__all__ = ('PluginState',)



from typing import Callable, TypeVar, Generic, Optional, Any

from amino import List, Nil, Map, __, Maybe
from amino.dat import Dat

from ribosome.config.component import Component
from ribosome.compute.program import Program
from ribosome.config.basic_config import NoData, BasicConfig
from ribosome.components.internal.config import internal
from ribosome.rpc.api import RpcProgram

A = TypeVar('A')
D = TypeVar('D')
CC = TypeVar('CC')
G = TypeVar('G')


class Config(Generic[D, CC], Dat['Config[D, CC]']):

    @staticmethod
    def from_opt(data: Map) -> 'Config[D, CC]':
        return Config.cons(data.lift('name') | 'no name in json', data.lift('prefix') | None)

    @staticmethod
    def cons(
            name: str,
            prefix: Optional[str]=None,
            state_ctor: Callable[[], D]=None,
            components: Map[str, Component[Any, CC]]=Map(),
            rpc: List[RpcProgram]=Nil,
            core_components: List[str]=Nil,
            default_components: List[str]=Nil,
            init: Program=None,
            internal_component: bool=True,
    ) -> 'Config[D, CC]':
        basic = BasicConfig.cons(
            name,
            prefix,
            state_ctor,
            core_components,
            default_components,
            internal_component,
        )
        return Config(
            basic,
            components + ('internal', internal) if internal_component else components,
            rpc,
            Maybe.optional(init),
        )

    def __init__(
            self,
            basic: BasicConfig[D],
            components: Map[str, Component[Any, CC]],
            rpc: List[RpcProgram],
            init: Maybe[Program],
    ) -> None:
        self.basic = basic
        self.components = components
        self.rpc = rpc
        self.init = init


__all__ = ('Config', 'NoData')



from typing import Callable, TypeVar, Generic

from amino import Dat

R = TypeVar('R')
S = TypeVar('S')


class ProgWrappers(Generic[R, S], Dat['TransWrappers[R, S]']):

    @staticmethod
    def id() -> 'ProgWrappers[R, R]':
        return ProgWrappers(lambda a: a, lambda a, b: b)

    def __init__(self, get: Callable[[R], S], put: Callable[[R, S], R]) -> None:
        self.get = get
        self.put = put


__all__ = ('ProgWrappers',)



from typing import TypeVar, Callable, Generic, Any, Type, Union

from amino import List, Either, _, Nil, Maybe, Boolean, __, Map, do, Do
from amino.dat import Dat

from ribosome.nvim.io.state import NS
from ribosome.data.mapping import Mappings
from ribosome.rpc.api import RpcProgram
from ribosome.compute.program import Program
from ribosome.util.doc.data import DocBlock

D = TypeVar('D')
CC = TypeVar('CC')
CD = TypeVar('CD')


class NoComponentData(Dat['NoComponentData']):
    pass


class ComponentData(Generic[D, CD], Dat['ComponentData[D, CD]']):

    def __init__(self, main: D, comp: CD) -> None:
        self.main = main
        self.comp = comp


def comp_data() -> NS[ComponentData[D, CD], CD]:
    return NS.inspect(_.comp)


@do(Maybe[Callable[[], CD]])
def infer_state_ctor(state_type: Union[None, Type[CD]]) -> Do:
    tpe = yield Maybe.optional(state_type)
    return getattr(tpe, 'cons', state_type)


class Component(Generic[CD, CC], Dat['Component[CD, CC]']):

    @staticmethod
    def cons(
            name: str,
            rpc: List[RpcProgram]=Nil,
            state_type: Type[CD]=None,
            state_ctor: Callable[[], CD]=None,
            config: CC=None,
            mappings: Mappings=None,
            help: DocBlock=None,
    ) -> 'Component[CD, CC]':
        return Component(
            name,
            rpc,
            state_type or NoComponentData,
            Maybe.optional(state_ctor).o(infer_state_ctor(state_type)),
            Maybe.optional(config),
            mappings or Mappings.cons(),
            Maybe.optional(help),
        )

    def __init__(
            self,
            name: str,
            rpc: List[RpcProgram],
            state_type: Type[CD],
            state_ctor: Maybe[Callable[[], CD]],
            config: Maybe[CC],
            mappings: Mappings,
            help: Maybe[DocBlock],
    ) -> None:
        self.name = name
        self.rpc = rpc
        self.state_type = state_type
        self.state_ctor = state_ctor
        self.config = config
        self.mappings = mappings
        self.help = help

    def handler_by_name(self, name: str) -> Either[str, Program]:
        return self.handlers.find(_.name == name).to_either(f'component `{self.name}` has no program `{name}`')

    def contains(self, prog: Program) -> Boolean:
        return self.rpc.contains(lambda a: a.program == prog)


class Components(Generic[CC], Dat['Components']):

    @staticmethod
    def cons(
            all: List[Component[Any, CC]]=Nil,
    ) -> 'Components[CC]':
        return Components(all)

    def __init__(self, all: List[Component[Any, CC]]) -> None:
        self.all = all

    def by_name(self, name: str) -> Either[str, Component[CD, CC]]:
        return self.all.find(_.name == name).to_either(f'no component named {name}')

    def by_type(self, tpe: type) -> Either[str, Component[CD, CC]]:
        return self.all.find(_.state_type == tpe).to_either(f'no component with state type {tpe}')

    @property
    def config(self) -> List[CC]:
        return self.all.collect(_.config)

    def for_program(self, prog: Program) -> Maybe[Component]:
        return self.all.find(__.contains(prog))


class ComponentConfig(Dat['ComponentConfig']):

    @staticmethod
    def cons(
            available: Map[str, Component[Any, CC]],
    ) -> 'ComponentConfig':
        return ComponentConfig(
            available,
        )

    def __init__(self, available: Map[str, Component[Any, CC]]) -> None:
        self.available = available


__all__ = ('Component', 'Components')



from typing import TypeVar

from amino import List, IO, Dat, do, Do
from amino.case import Case
from amino.state import State
from amino.lenses.lens import lens

from ribosome.util.doc.data import (DocLine, DocCompiler, DocCat, DocFragment, DocString, DocBlock, DocMeta, NoMeta,
                                    Headline, Code, CustomDocMeta, Link, AnchorType, GeneralAnchor, VariableAnchor,
                                    RpcAnchor, MappingAnchor)
from ribosome.util.doc.format import CompilerConfig, compile_anchor

A = TypeVar('A')


class MarkdownCompilerConfig(Dat['MarkdownCompilerConfig']):

    @staticmethod
    def cons(
            basic: CompilerConfig,
    ) -> 'MarkdownCompilerConfig':
        return MarkdownCompilerConfig(
            basic,
        )

    def __init__(self, basic: CompilerConfig) -> None:
        self.basic = basic


class postproc_anchor(Case[AnchorType, str], alg=AnchorType):

    def __init__(self, text: str) -> None:
        self.text = text

    def variable(self, a: VariableAnchor) -> str:
        return f'`{self.text}`'

    def rpc(self, a: RpcAnchor) -> str:
        return f'`{self.text}`'

    def mapping(self, a: MappingAnchor) -> str:
        return f'`{self.text}`'

    def general(self, a: GeneralAnchor) -> str:
        return self.text


class compile_string(Case[DocMeta[A], State[MarkdownCompilerConfig, List[str]]], alg=DocMeta):

    def __init__(self, string: DocString[A]) -> None:
        self.string = string

    def no(self, a: NoMeta[A]) -> State[MarkdownCompilerConfig, List[str]]:
        return State.pure(List(self.string.text))

    @do(State[MarkdownCompilerConfig, List[str]])
    def headline(self, a: Headline[A]) -> Do:
        hashes = '#' * a.level
        anchor = yield (
            a.anchor
            .map(lambda b: compile_anchor(b).map(lambda c: postproc_anchor(c)(b.tpe)))
            .get_or(lambda: State.pure(f'{self.string.text}'))
            .zoom(lens.basic)
        )
        return List(f'{hashes} {anchor}', '')

    def code(self, a: Code[A]) -> State[MarkdownCompilerConfig, List[str]]:
        return State.pure(List(f'`{self.string.text}`'))

    def link(self, a: Link[A]) -> State[MarkdownCompilerConfig, List[str]]:
        return State.pure(List(f'`[{a.desc}]({self.string.text})`'))

    def custom(self, a: CustomDocMeta[A]) -> State[MarkdownCompilerConfig, List[str]]:
        return State.pure(List(self.string.text))


class compile_fragment(Case[DocFragment[A], List[str]], alg=DocFragment):

    @do(State[MarkdownCompilerConfig, List[str]])
    def cat(self, frag: DocCat[A]) -> Do:
        parts = yield frag.fragments.flat_traverse(self, State)
        return List(parts.join_tokens)

    def string(self, frag: DocString[A]) -> State[MarkdownCompilerConfig, List[str]]:
        return compile_string(frag)(frag.meta)


def compile_line(line: DocLine[A]) -> State[MarkdownCompilerConfig, List[str]]:
    return compile_fragment.match(line.data)


def compile_block(block: DocBlock[A]) -> State[MarkdownCompilerConfig, List[str]]:
    return block.lines.flat_traverse(compile_line, State)


def compile_markdown(blocks: List[DocBlock[A]], conf: MarkdownCompilerConfig) -> IO[List[str]]:
    return IO.pure(blocks.flat_traverse(compile_block, State).run_a(conf).value)


def markdown_compiler(conf: MarkdownCompilerConfig) -> DocCompiler[A, MarkdownCompilerConfig]:
    return DocCompiler(compile_markdown, conf)


__all__ = ('compile_markdown', 'markdown_compiler',)



import time
from datetime import datetime
from typing import Callable, Any, Union, TypeVar, Generic

from amino.test.spec import default_timeout
from amino import List

from kallikrein import Expectation, kf
from kallikrein.matcher import BoundMatcher
from kallikrein.matchers.length import have_length
from kallikrein.matchers.comparison import greater
from kallikrein.matchers import contain, equal
from kallikrein.matchers.lines import have_lines
from kallikrein.matchers.maybe import be_just
from kallikrein.matchers.either import be_right

from ribosome.test.klk.expectable import kn
from ribosome.nvim.api.variable import variable_prefixed_raw, variable_raw
from ribosome.nvim.api.exists import command_exists


def later_f(exp: Callable[[], Expectation], timeout: float=None, intval: float=0.1) -> Expectation:
    to = default_timeout if timeout is None else timeout
    start = datetime.now()
    while (not exp().unsafe_eval and (datetime.now() - start).total_seconds() < to):
        time.sleep(intval)
    exp().fatal_eval()
    return exp()


def later(exp: Expectation, timeout: float=None, intval: float=0.1) -> Expectation:
    return later_f(lambda: exp, timeout, intval)


# class VimIntegrationKlkHelpers(VimIntegrationSpecI):

#     def _wait_for(self, pred: Callable[..., bool], *a: Any, **kw: Any) -> Expectation:
#         return later(kf(pred, *a, **kw).true)

#     def _list_line(self, log: Callable[[], List[str]], index: int, matcher: BoundMatcher) -> Expectation:
#         minlen = index if index >= 0 else abs(index + 1)
#         return later(
#             kf(log).must(have_length(greater(minlen))) &
#             kf(lambda: log().lift(index)).must(matcher)
#         )

#     def _log_line(self, index: int, matcher: BoundMatcher) -> Expectation:
#         return self._list_line(lambda: self._log_out, index, matcher)

#     def _log_contains(self, target: Union[str, BoundMatcher[str]]) -> Expectation:
#         matcher = target if isinstance(target, BoundMatcher) else equal(target)
#         return later(kf(lambda: self._log_out).must(contain(matcher)))

#     def _log_equals(self, lines: Union[str, List[str]]) -> Expectation:
#         return later(kf(lambda: self._log_out).must(have_lines(lines)))

#     def _buffer_content(self, data: Union[str, List[str]]) -> Expectation:
#         return later(self.contentkf.must(have_lines(data)))

#     def _buffer_length(self, length: int) -> Expectation:
#         return later(self.contentkf.must(have_length(length)))

#     def _buffer_line(self, index: int, matcher: BoundMatcher) -> Expectation:
#         return self._list_line(lambda: self._buffer_out, index, be_just(matcher))

#     def _buffer_contains(self, line: str) -> Expectation:
#         return later(kf(lambda: self._buffer_out).must(contain(line)))

#     def _buffer_option(self, name: str, value: Any) -> Expectation:
#         return later(kf(lambda: self.buffer.options(name)).must(be_right(value)))

#     def _buffer_name(self, name: str) -> Expectation:
#         return later(kf(lambda: self.buffer.name) == name)

#     def _buffer_count(self, count: int) -> Expectation:
#         return later(kf(lambda: self.vim.buffers).must(have_length(count)))

#     def _window_height(self, height: int) -> Expectation:
#         return later(kf(lambda: self.vim.window.height).must(contain(height)))

#     def _window_count(self, count: int) -> Expectation:
#         return later(kf(lambda: self.vim.windows).must(have_length(count)))

#     def _cursor_line(self, lnum: int) -> Expectation:
#         return later(kf(lambda: self.vim.window.line).must(contain(lnum)))

#     def _messages_contain(self, line: str) -> Expectation:
#         return later(kf(lambda: self.vim.messages).must(contain(line)))

#     def seen_program(self, name: str, **kw) -> Expectation:
#         return later(kf(self.program_log).must(contain(name)), **kw)

#     def var_is(self, name: str, value: Any) -> Expectation:
#         return kn(self.vim, variable_raw, name).must(contain(be_right(value)))

#     def var_becomes(self, name: str, value: Any) -> Expectation:
#         return later(self.var_is(name, value))

#     def pvar_is(self, name: str, value: Any) -> Expectation:
#         return kn(self.vim, variable_prefixed_raw, name).must(contain(be_right(value)))

#     def pvar_becomes(self, name: str, value: Any) -> Expectation:
#         return later(self.pvar_is(name, value))

#     def pvar_becomes_map(self, name: str, value: Any, f: Callable[[Any], Any]) -> Expectation:
#         return later(kf(lambda: self.vim.vars.p(name).map(f)).must(be_right(value)))
#         return self._wait_for(lambda: self.vim.vars.p(name).map(f).contains(value))

#     def command_exists(self, name: str, **kw) -> Expectation:
#         return later(kn(self.vim, command_exists, name).must(contain(True)), **kw)

#     def command_exists_not(self, name: str, **kw: Any) -> Expectation:
#         return later(kn(self.vim, command_exists, name).must(contain(False)), **kw)



from amino import Dat, List, Nil

from ribosome.nvim.syntax.expr import SyntaxItem, Highlight, HiLink


class Syntax(Dat['Syntax']):

    @staticmethod
    def cons(
            syntax: List[SyntaxItem]=Nil,
            highlight: List[Highlight]=Nil,
            links: List[HiLink]=Nil,
    ) -> 'Syntax':
        return Syntax(
            syntax,
            highlight,
            links,
        )

    def __init__(self, syntax: List[SyntaxItem], highlight: List[Highlight], links: List[HiLink]) -> None:
        self.syntax = syntax
        self.highlight = highlight
        self.links = links


__all__ = ('Syntax',)



from __future__ import annotations
from typing import Any, Callable, Optional

from ribosome.config.component import Component

from amino import Map, Lists, Dat, Path, List, Nil, Maybe
from amino.test import temp_dir
from ribosome.config.config import Config
from ribosome.compute.program import Program
from ribosome.rpc.api import RpcProgram, rpc
from ribosome.nvim.io.compute import NvimIO
from ribosome.compute.interpret import ProgIOInterpreter
from ribosome.nvim.io.api import N
from ribosome.test.request import Handler, no_handler


default_config_name = 'spec'
no_pre = lambda: N.unit


class TestConfig(Dat['TestConfig']):

    @staticmethod
    def cons(
            config: Config,
            pre: Callable[[], NvimIO[None]]=None,
            log_dir: Path=None,
            log_file: Path=None,
            components: List[str]=Nil,
            io_interpreter: ProgIOInterpreter=None,
            logger: Program[None]=None,
            vars: Map[str, Any]=Map(),
            request_handler: Handler=no_handler,
            function_handler: Handler=no_handler,
            command_handler: Handler=no_handler,
            autostart: bool=True,
            config_path: str=None,
            tmux_socket: str=None,
    ) -> 'TestConfig':
        ld = log_dir or temp_dir('log')
        lf = log_file or ld / config.basic.name
        return TestConfig(
            config,
            pre or no_pre,
            ld,
            lf,
            components,
            io_interpreter,
            logger,
            vars,
            request_handler,
            function_handler,
            command_handler,
            autostart,
            Maybe.optional(config_path),
            Maybe.optional(tmux_socket),
        )

    def __init__(
            self,
            config: Config,
            pre: Callable[[], NvimIO[None]],
            log_dir: Path,
            log_file: Path,
            components: List[str],
            io_interpreter: Optional[ProgIOInterpreter],
            logger: Optional[Program[None]],
            vars: Map[str, Any],
            request_handler: Handler,
            function_handler: Handler,
            command_handler: Handler,
            autostart: bool,
            config_path: Maybe[str],
            tmux_socket: Maybe[str],
    ) -> None:
        self.config = config
        self.pre = pre
        self.log_dir = log_dir
        self.log_file = log_file
        self.components = components
        self.io_interpreter = io_interpreter
        self.logger = logger
        self.vars = vars
        self.request_handler = request_handler
        self.function_handler = function_handler
        self.command_handler = command_handler
        self.autostart = autostart
        self.config_path = config_path
        self.tmux_socket = tmux_socket

    def with_vars(self, **kw: Any) -> 'TestConfig':
        return self.copy(vars=self.vars ** Map(kw))


def spec_config(*rpc: RpcProgram) -> Config:
    component: Component = Component.cons(
        'main',
        rpc=Lists.wrap(rpc)
    )
    return Config.cons(
        name=default_config_name,
        prefix=default_config_name,
        components=Map(main=component),
        internal_component=False,
    )


def single_prog_config(prog: Program, **kw: Any) -> Config:
    return spec_config(rpc.write(prog).conf(**kw))


__all__ = ('single_prog_config', 'spec_config')



from typing import TypeVar
from types import ModuleType

from amino.logging import module_log
from amino import List, IO, do, Do, Path
from amino.io import IOException

from ribosome.util.doc.data import StaticDoc, DocCompiler
from ribosome.util.doc.markdown import markdown_compiler, MarkdownCompilerConfig
from ribosome.util.doc.vim import vim_compiler, VimCompilerConfig
from ribosome.util.doc.generate import generate_plugin_doc
from ribosome.util.doc.format import CompilerConfig

log = module_log()
A = TypeVar('A')
B = TypeVar('B')


def report_error(error: IOException) -> None:
    log.error(f'failed to generate doc: {error}')


@do(IO[None])
def run(
        components: str,
        settings: List[ModuleType],
        compiler: DocCompiler[A, B],
        outfile: Path,
        static: StaticDoc,
) -> Do:
    text = yield generate_plugin_doc(components, settings, static, compiler)
    yield IO.delay(outfile.write_text, text.join_lines)


@do(IO[None])
def run_default(
        components: str,
        settings: List[ModuleType],
        pkg_dir: Path,
        name: str,
        prefix: str,
        static: StaticDoc,
) -> Do:
    vim_conf = VimCompilerConfig.cons(CompilerConfig.cons(name, prefix))
    md_conf = MarkdownCompilerConfig.cons(CompilerConfig.cons(name, prefix))
    yield run(components, settings, markdown_compiler(md_conf), pkg_dir / 'README.md', static)
    yield run(components, settings, vim_compiler(vim_conf), pkg_dir / 'data' / 'runtime' / f'{name}.txt', static)


def write_default_docs(
        components: str,
        settings: List[ModuleType],
        pkg_dir: Path,
        name: str,
        prefix: str,
        static: StaticDoc,
) -> None:
    run_default(components, settings, pkg_dir, name, prefix, static).attempt.lmap(report_error)

__all__ = ('write_default_docs',)



import abc
from typing import Generic, TypeVar, Tuple
from traceback import FrameSummary

from amino import ADT, Either, Right, Left, Dat, Nil
from amino.util.trace import cframe
from amino.state import State

from ribosome.nvim.io.trace import NvimIOException

A = TypeVar('A')
B = TypeVar('B')


class NResult(Generic[A], ADT['NResult[A]']):

    @abc.abstractproperty
    def to_either(self) -> Either[Exception, A]:
        ...


class NSuccess(Generic[A], NResult[A]):

    def __init__(self, value: A) -> None:
        self.value = value

    @property
    def to_either(self) -> Either[Exception, A]:
        return Right(self.value)


class NError(Generic[A], NResult[A]):

    def __init__(self, error: str) -> None:
        self.error = error

    @property
    def to_either(self) -> Either[Exception, A]:
        return Left(Exception(self.error))


class NFatal(Generic[A], NResult[A]):

    def __init__(self, exception: Exception) -> None:
        self.exception = exception

    @property
    def to_either(self) -> Either[Exception, A]:
        return Left(self.exception)


class Thunk(Generic[A, B], Dat['Thunk[A, B]']):

    @staticmethod
    def cons(thunk: State[A, B], frame: FrameSummary=None) -> 'Thunk[A, B]':
        return Thunk(thunk, frame or cframe())

    def __init__(self, thunk: State[A, B], frame: FrameSummary) -> None:
        self.thunk = thunk
        self.frame = frame


def eval_thunk(resource: A, thunk: Thunk[A, B]) -> Tuple[A, B]:
    try:
        return thunk.thunk.run(resource).value
    except NvimIOException as e:
        raise e
    except Exception as e:
        raise NvimIOException('', Nil, e, thunk.frame)


__all__ = ('NResult', 'NSuccess', 'NError', 'NFatal', 'Thunk', 'eval_thunk')



from typing import Callable, TypeVar, Any

from amino import do, Do, Either, Right, Left

from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.api.function import nvim_call_cons_strict, nvim_call_function
from ribosome.nvim.api.util import nvimio_repeat_timeout, cons_checked_e
from ribosome.nvim.api.command import nvim_command
from ribosome.nvim.io.api import N

A = TypeVar('A')


def check_exists_bool(a: Any) -> Either[str, bool]:
    return Right(a != 0) if a in [0, 1, 2] else Left(f'nvim result is not an exists-boolean: {a}')


cons_decode_exists_bool = cons_checked_e(int, check_exists_bool)


def nvim_exists(target: str) -> NvimIO[bool]:
    return nvim_call_cons_strict(cons_decode_exists_bool, 'exists', target)


def wait_until_valid(
        name: str,
        check: Callable[[str], NvimIO[bool]],
        timeout: float=30.,
        interval: float=.01,
        desc: str='satify condition',
) -> NvimIO[None]:
    return nvimio_repeat_timeout(
        lambda: check(name),
        lambda a: a,
        f'{name} did not {desc} within {timeout} seconds',
        timeout,
        interval,
    )


def function_exists(name: str) -> NvimIO[bool]:
    return nvim_exists(f'*{name}')


@do(NvimIO[bool])
def function_exists_not(name: str) -> Do:
    exists = yield function_exists(name)
    return not exists


def command_exists(name: str) -> NvimIO[bool]:
    return nvim_exists(f':{name}')


@do(NvimIO[bool])
def command_exists_not(name: str) -> Do:
    exists = yield command_exists(name)
    return not exists


def wait_for_function(name: str, timeout: int=30, **kw: Any) -> NvimIO[None]:
    return wait_until_valid(name, function_exists, timeout=timeout, desc='appear', **kw)


def wait_for_function_undef(name: str, timeout: int=30) -> NvimIO[None]:
    return wait_until_valid(name, function_exists_not, timeout)


def wait_for_command(name: str, timeout: int=30) -> NvimIO[None]:
    return wait_until_valid(name, command_exists, timeout=timeout, desc='appear')


@do(NvimIO[A])
def command_once_defined(name: str, *args: str, timeout: int=30, verbose: bool=False) -> Do:
    yield wait_for_command(name, timeout=timeout)
    yield nvim_command(name, *args, verbose=verbose)


@do(NvimIO[A])
def call_once_defined(name: str, *args: str, timeout: int=10) -> Do:
    yield wait_for_function(name, timeout=timeout)
    yield nvim_call_function(name, *args)


def wait_until_function_produces(
        target: Any,
        name: str,
        *args: str,
        timeout: int=10,
        interval: float=None,
) -> NvimIO[A]:
    return nvimio_repeat_timeout(
        lambda: N.recover_failure(nvim_call_function(name, *args), N.pure),
        lambda a: a == target,
        f'{name} did not produce `{target}` within {timeout} seconds',
        timeout,
        interval,
    )


__all__ = ('nvim_exists', 'wait_until_valid', 'function_exists', 'command_exists', 'wait_for_function',
           'wait_for_command', 'command_once_defined', 'call_once_defined', 'function_exists_not',
           'wait_for_function_undef', 'command_exists_not', 'wait_until_function_produces',)
